/*****************************************************************************

        FilterResize.cpp
        Author: Laurent de Soras, 2011

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if defined (_MSC_VER)
	#pragma warning (1 : 4130 4223 4705 4706)
	#pragma warning (4 : 4355 4786 4800)
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include "conc/def.h"
#include "ContFirInterface.h"
#include "FilterResize.h"
#include "fnc.h"
#include "Scaler.h"
#include "ToolsSse2.h"

#include <algorithm>

#include <cassert>
#include <cmath>
#include <cstring>



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



FilterResize::FilterResize (int src_width, int src_height, int dst_width, int dst_height, double win_x, double win_y, double win_w, double win_h, ContFirInterface &kernel_fnc_h, ContFirInterface &kernel_fnc_v, double kernel_scale_h, double kernel_scale_v, bool norm_flag, double norm_val_h, double norm_val_v, double center_pos_h, double center_pos_v, bool sse2_flag)
:	_avstp (AvstpWrapper::use_instance ())
,	_task_rsz_pool ()
/*,	_src_size ()
,	_dst_size ()
,	_win_pos ()
,	_win_size ()
,	_kernel_scale ()
,	_kernel_force_flag ()
,	_kernel_ptr_arr ()*/
,	_norm_flag (norm_flag)
/*,	_norm_val ()
,	_center_pos ()*/
,	_sse2_flag (sse2_flag)
,	_pool ()
,	_factory_aptr ()
/*,	_crop_pos ()
,	_crop_size ()
,	_scaler_aptr ()*/
,	_conv_lsb_msb_to_float_ptr (&ThisType::conv_lsb_msb_to_float_cpp)
,	_conv_float_to_lsb_msb_ptr (&ThisType::conv_float_to_lsb_msb_cpp)
/*,	_resize_flag ()
,	_roadmap ()
,	_tile_size_dst ()*/
,	_nbr_passes (0)
,	_buffer_flag (false)
{
	assert (src_width > 0);
	assert (src_height > 0);
	assert (dst_width > 0);
	assert (dst_height > 0);
	assert (win_w > 0);
	assert (win_h > 0);
	assert (kernel_scale_h != 0);
	assert (kernel_scale_v != 0);
	assert (&kernel_fnc_h != 0);
	assert (&kernel_fnc_v != 0);

	_src_size [Dir_H]          = src_width;
	_src_size [Dir_V]          = src_height;
	_dst_size [Dir_H]          = dst_width;
	_dst_size [Dir_V]          = dst_height;
	_win_pos [Dir_H]           = win_x;
	_win_pos [Dir_V]           = win_y;
	_win_size [Dir_H]          = win_w;
	_win_size [Dir_V]          = win_h;
	_kernel_scale [Dir_H]      = fabs (kernel_scale_h);
	_kernel_scale [Dir_V]      = fabs (kernel_scale_v);
	_kernel_force_flag [Dir_H] = (kernel_scale_h < 0);
	_kernel_force_flag [Dir_V] = (kernel_scale_v < 0);
	_center_pos [Dir_H]        = center_pos_h;
	_center_pos [Dir_V]        = center_pos_v;
	_crop_pos [Dir_H]          = 0;
	_crop_pos [Dir_V]          = 0;
	_crop_size [Dir_H]         = src_width;
	_crop_size [Dir_V]         = src_height;
	_kernel_ptr_arr [Dir_H]    = &kernel_fnc_h;
	_kernel_ptr_arr [Dir_V]    = &kernel_fnc_v;
	_norm_val [Dir_H]          = norm_val_h;
	_norm_val [Dir_V]          = norm_val_v;

	if (_sse2_flag)
	{
		_conv_lsb_msb_to_float_ptr = &ThisType::conv_lsb_msb_to_float_sse;
		_conv_float_to_lsb_msb_ptr = &ThisType::conv_float_to_lsb_msb_sse;
	}

	for (int dir = 0; dir < Dir_NBR_ELT; ++dir)
	{
		_resize_flag [dir] = (
			   _win_pos [dir] < 0
		   || _win_pos [dir] + _win_size [dir] > _src_size [dir]
		   || fabs (_win_pos [dir] - round (_win_pos [dir])) > 1e-5
		   || _win_size [dir] != _dst_size [dir]
		   || _kernel_scale [dir] != 1.0
		   || _kernel_force_flag [dir]
		);

		if (_resize_flag [dir])
		{
			Scaler::eval_req_src_area (
				_crop_pos [dir], _crop_size [dir],
				_src_size [dir], _dst_size [dir],
				_win_pos [dir], _win_size [dir],
				*(_kernel_ptr_arr [dir]), _kernel_scale [dir], _center_pos [dir]
			);

			_scaler_aptr [dir] = std::auto_ptr <Scaler> (new Scaler (
				_crop_size [dir], _dst_size [dir],
				_win_pos [dir] - _crop_pos [dir], _win_size [dir],
				*(_kernel_ptr_arr [dir]), _kernel_scale [dir],
				_norm_flag, _norm_val [dir], _center_pos [dir], _sse2_flag
			));
		}
		else
		{
			_crop_pos [dir]  = round_int (_win_pos [dir]);
			_crop_size [dir] = round_int (_win_size [dir]);
		}
	}

	// Finds the scaling order
	const double		r_h = _dst_size [Dir_H] / _win_size [Dir_H];
	const double		r_v = _dst_size [Dir_V] / _win_size [Dir_V];
	bool					vert_last_flag =
		   (r_v < 1 && r_v < r_h)	// Vertical downscale has the highest reduction factor
		|| (r_v > 1 && r_h > 1);	// Upscale in both direction: avoids using a buffer to store the final picture.

	// Special case for extreme deformations
	vert_last_flag = (vert_last_flag ||   (r_v / r_h > 8 && r_v > 4));
	vert_last_flag = (vert_last_flag && ! (r_h / r_v > 8 && r_h > 4));

	// Builds a roadmap
	_buffer_flag = false;
	int					rm_pos = 0;
	if (_resize_flag [Dir_V] && ! vert_last_flag)
	{
		_roadmap [rm_pos] = PassType_RESIZE;
		++ rm_pos;
	}
	if (_resize_flag [Dir_H])
	{
		_buffer_flag = true;
		_roadmap [rm_pos    ] = PassType_TRANSPOSE;
		_roadmap [rm_pos + 1] = PassType_RESIZE;
		_roadmap [rm_pos + 2] = PassType_TRANSPOSE;
		rm_pos += 3;
	}
	if (_resize_flag [Dir_V] && vert_last_flag)
	{
		_roadmap [rm_pos] = PassType_RESIZE;
		++ rm_pos;
	}
	assert (rm_pos <= MAX_NBR_PASSES);

	_nbr_passes = rm_pos;

	while (rm_pos < MAX_NBR_PASSES)
	{
		_roadmap [rm_pos] = PassType_NONE;
		++ rm_pos;
	}

	// Computes the tile size (if required)
	if (_buffer_flag)
	{
		const int		tile_dst_min_w = std::min (_dst_size [Dir_H], int (Scaler::SRC_ALIGN));
		const int		tile_dst_min_h = 1;

		int				tile_src_min_w;
		int				tile_src_min_h;
		compute_req_src_tile_size (
			tile_src_min_w,
			tile_src_min_h,
			tile_dst_min_w,
			tile_dst_min_h
		);
		tile_src_min_w = (tile_src_min_w + Scaler::SRC_ALIGN - 1) & -Scaler::SRC_ALIGN;
		tile_src_min_h = (tile_src_min_h + Scaler::SRC_ALIGN - 1) & -Scaler::SRC_ALIGN;

		// First computes a well-balanced source size
		int				tile_src_w = tile_src_min_w;
		int				tile_src_h = tile_src_min_h;
		const double	a = 1;
		const double	b = tile_src_min_w + tile_src_min_h;
		const double	c = tile_src_min_w * tile_src_min_h - MAX_BUF_SIZE;
		const double	d = b * b - 4 * a * c;
		const int		l = round_int ((sqrt (d) - b) / (2 * a));

		tile_src_w = (l + tile_src_min_w);
		bool				ok_flag = false;
		do
		{
			const int		old_tile_src_w = tile_src_w;
			tile_src_w = std::min (tile_src_w, _crop_size [Dir_H]);
			if (_resize_flag [Dir_V])
			{
				tile_src_w = (tile_src_w + Scaler::SRC_ALIGN - 1) & -Scaler::SRC_ALIGN;
			}
			tile_src_h = MAX_BUF_SIZE / tile_src_w;
			tile_src_h = std::min (tile_src_h, _crop_size [Dir_V] + Scaler::SRC_ALIGN - 1);
			tile_src_h &= -Scaler::SRC_ALIGN;
			ok_flag = (tile_src_h >= std::min (tile_src_min_h, _crop_size [Dir_V]));
			if (! ok_flag)
			{
				tile_src_w = old_tile_src_w / 2;
				if (   tile_src_w < tile_src_min_w
				    || tile_src_w >= old_tile_src_w)
				{
					assert (false);
					throw ("Dither resizer: cannot resize with these parameters.");
				}
			}
		}
		while (! ok_flag);

		// Computes the equivalent destination size
		int			tile_dst_w = tile_src_w;
		if (_resize_flag [Dir_H])
		{
			tile_dst_w = Scaler::eval_lower_bound_of_dst_tile_height (
				tile_src_w,
				_dst_size [Dir_H],
				_win_size [Dir_H],
				*(_kernel_ptr_arr [Dir_H]),
				_kernel_scale [Dir_H],
				_crop_size [Dir_H]
			);
			tile_dst_w = std::max (tile_dst_w, tile_dst_min_w);
		}

		int			tile_dst_h = tile_src_h;
		if (_resize_flag [Dir_V])
		{
			tile_dst_h = Scaler::eval_lower_bound_of_dst_tile_height (
				tile_src_h,
				_dst_size [Dir_V],
				_win_size [Dir_V],
				*(_kernel_ptr_arr [Dir_V]),
				_kernel_scale [Dir_V],
				_crop_size [Dir_V]
			);
			tile_dst_h = std::max (tile_dst_h, tile_dst_min_h);
		}

		// Limits the destination size
		if (! vert_last_flag)	// If we use a buffer for the last step
		{
			const double	area_final = double (tile_dst_w) * double (tile_dst_h);
			assert (area_final > 0);
			const double	fix_final = sqrt (MAX_BUF_SIZE / area_final);
			if (fix_final < 1)
			{
				tile_dst_w = floor_int (tile_dst_w * fix_final) & -Scaler::SRC_ALIGN;
				tile_dst_h = floor_int (tile_dst_h * fix_final) & -Scaler::SRC_ALIGN;
			}
		}

		// Limits the temporary size
		if (vert_last_flag)
		{
			const int		stride =
				(tile_dst_w + Scaler::SRC_ALIGN - 1) & -Scaler::SRC_ALIGN;
			const double	area_tmp = stride * tile_src_h;
			assert (area_tmp > 0);
			const double	fix_tmp = MAX_BUF_SIZE / area_tmp;
			if (fix_tmp < 1)
			{
				tile_dst_w = floor_int (tile_dst_w * fix_tmp);
			}
//			tile_dst_h &= -Scaler::SRC_ALIGN;
		}
		else
		{
			const int		stride =
				(tile_src_w + Scaler::SRC_ALIGN - 1) & -Scaler::SRC_ALIGN;
			const double	area_tmp = stride * tile_dst_h;
			assert (area_tmp > 0);
			const double	fix_tmp = MAX_BUF_SIZE / area_tmp;
			if (fix_tmp < 1)
			{
				tile_dst_h = floor_int (tile_dst_h * fix_tmp) & -Scaler::SRC_ALIGN;
			}
//			tile_dst_w &= -Scaler::SRC_ALIGN;
		}

		// For the last step, we need to ensure that the window width is well
		// aligned, otherwise we might overflow the end of the lines by
		// starting the rightmost tile on an unaligned column.
		if (tile_dst_w < _dst_size [Dir_H])
		{
			tile_dst_w &= -Scaler::SRC_ALIGN;
		}

		if (tile_dst_w <= 0 || tile_dst_h <= 0)
		{
			throw std::runtime_error (
				"Dither_resize16: "
				"resizing ratio too low or kernel support too high."
			);
		}

		tile_dst_w = std::max (tile_dst_w, int (Scaler::SRC_ALIGN));
		tile_dst_h = std::max (tile_dst_h, 1);

		_tile_size_dst [Dir_H] = tile_dst_w;
		_tile_size_dst [Dir_V] = tile_dst_h;

		_factory_aptr = std::auto_ptr <ResizeDataFactory> (
			new ResizeDataFactory (MAX_BUF_SIZE, 1)
		);
		_pool.set_factory (*_factory_aptr);
	}

	// ! buffer_flag
	else
	{
		_tile_size_dst [Dir_H] = _dst_size [Dir_H];
		_tile_size_dst [Dir_V] = _dst_size [Dir_V];
	}

}



void	FilterResize::process_plane (uint8_t *dst_msb_ptr, uint8_t *dst_lsb_ptr, const uint8_t *src_msb_ptr, const uint8_t *src_lsb_ptr, int stride_dst, int stride_src)
{
	assert (dst_msb_ptr != 0);
	assert (dst_lsb_ptr != 0);
	assert (src_msb_ptr != 0);
	assert (src_lsb_ptr != 0);
	assert (stride_dst > 0);
	assert (stride_src > 0);

	if (_nbr_passes <= 0)
	{
		process_plane_bypass (
			dst_msb_ptr, dst_lsb_ptr,
			src_msb_ptr, src_lsb_ptr,
			stride_dst, stride_src
		);
	}
	else
	{
		process_plane_normal (
			dst_msb_ptr, dst_lsb_ptr,
			src_msb_ptr, src_lsb_ptr,
			stride_dst, stride_src
		);
	}
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



void	FilterResize::process_plane_bypass (uint8_t *dst_msb_ptr, uint8_t *dst_lsb_ptr, const uint8_t *src_msb_ptr, const uint8_t *src_lsb_ptr, int stride_dst, int stride_src)
{
	assert (_nbr_passes <= 0);
	assert (dst_msb_ptr != 0);
	assert (dst_lsb_ptr != 0);
	assert (src_msb_ptr != 0);
	assert (src_lsb_ptr != 0);
	assert (stride_dst > 0);
	assert (stride_src > 0);

	const int		offset_src =
		  round_int (_win_pos [Dir_V]) * stride_src
		+ round_int (_win_pos [Dir_H]);

	for (int y = 0; y < _dst_size [Dir_V]; ++y)
	{
		memcpy (
			dst_msb_ptr + y * stride_dst,
			src_msb_ptr + y * stride_src + offset_src,
			_dst_size [Dir_H]
		);
	}

	for (int y = 0; y < _dst_size [Dir_V]; ++y)
	{
		memcpy (
			dst_lsb_ptr + y * stride_dst,
			src_lsb_ptr + y * stride_src + offset_src,
			_dst_size [Dir_H]
		);
	}
}



void	FilterResize::process_plane_normal (uint8_t *dst_msb_ptr, uint8_t *dst_lsb_ptr, const uint8_t *src_msb_ptr, const uint8_t *src_lsb_ptr, int stride_dst, int stride_src)
{
	assert (_nbr_passes > 0);
	assert (dst_msb_ptr != 0);
	assert (dst_lsb_ptr != 0);
	assert (src_msb_ptr != 0);
	assert (src_lsb_ptr != 0);
	assert (stride_dst > 0);
	assert (stride_src > 0);

	avstp_TaskDispatcher *	task_dispatcher_ptr = _avstp.create_dispatcher ();

	const int		offset_crop =
		_crop_pos [Dir_V] * stride_src + _crop_pos [Dir_H];

	TaskRszGlobal	trg;
	trg._this_ptr    = this;
	trg._dst_msb_ptr = dst_msb_ptr;
	trg._dst_lsb_ptr = dst_lsb_ptr;
	trg._src_msb_ptr = src_msb_ptr;
	trg._src_lsb_ptr = src_lsb_ptr;
	trg._stride_dst  = stride_dst;
	trg._stride_src  = stride_src;
	trg._offset_crop = offset_crop;

	int				dst_beg [Dir_NBR_ELT]  = { 0, 0 };
	int				work_dst [Dir_NBR_ELT] = { 0, 0 };
	int				src_beg [Dir_NBR_ELT]  = { 0, 0 };
	int				src_end [Dir_NBR_ELT]  = { 0, 0 };

	for (dst_beg [Dir_V] = 0
	;	dst_beg [Dir_V] < _dst_size [Dir_V]
	;	dst_beg [Dir_V] += _tile_size_dst [Dir_V])
	{
		work_dst [Dir_V] = std::min (
			_tile_size_dst [Dir_V],
			_dst_size [Dir_V] - dst_beg [Dir_V]
		);

		src_beg [Dir_V] = 0;
		src_end [Dir_V] = 0;
		if (_resize_flag [Dir_V])
		{
			_scaler_aptr [Dir_V]->get_src_boundaries (
				src_beg [Dir_V],
				src_end [Dir_V],
				dst_beg [Dir_V],
				dst_beg [Dir_V] + work_dst [Dir_V]
			);
		}
		else
		{
			src_beg [Dir_V] = dst_beg [Dir_V];
			src_end [Dir_V] = src_beg [Dir_V] + work_dst [Dir_V];
		}

		for (dst_beg [Dir_H] = 0
		;	dst_beg [Dir_H] < _dst_size [Dir_H]
		;	dst_beg [Dir_H] += _tile_size_dst [Dir_H])
		{
			work_dst [Dir_H] = std::min (
				_tile_size_dst [Dir_H],
				_dst_size [Dir_H] - dst_beg [Dir_H]
			);

			src_beg [Dir_H] = 0;
			src_end [Dir_H] = 0;
			if (_resize_flag [Dir_H])
			{
				_scaler_aptr [Dir_H]->get_src_boundaries (
					src_beg [Dir_H],
					src_end [Dir_H],
					dst_beg [Dir_H],
					dst_beg [Dir_H] + work_dst [Dir_H]
				);
			}
			else
			{
				src_beg [Dir_H] = dst_beg [Dir_H];
				src_end [Dir_H] = src_beg [Dir_H] + work_dst [Dir_H];
			}

			// The cell will be returned to the pool by the task.
			TaskRszCell *	tr_cell_ptr = _task_rsz_pool.take_cell (true);
			if (tr_cell_ptr == 0)
			{
				throw std::runtime_error (
					"Dither_resize16: Cannot allocate task cell."
				);
			}

			TaskRsz &		tr = tr_cell_ptr->_val;
			tr._glob_data_ptr = &trg;
			for (int d = 0; d < Dir_NBR_ELT; ++d)
			{
				tr._dst_beg [d]  = dst_beg [d];
				tr._src_beg [d]  = src_beg [d];
				tr._src_end [d]  = src_end [d];
				tr._work_dst [d] = work_dst [d];
			}

			_avstp.enqueue_task (
				task_dispatcher_ptr,
				&redirect_task_resize,
				tr_cell_ptr
			);
		}	// for Dir_H
	}	// for Dir_V

	_avstp.wait_completion (task_dispatcher_ptr);

	// Done
	_avstp.destroy_dispatcher (task_dispatcher_ptr);
	task_dispatcher_ptr = 0;
}



void	FilterResize::process_tile (TaskRszCell &tr_cell)
{
	assert (&tr_cell != 0);

#if (conc_WORD_SIZE == 32)
	// We can arrive here from any thread and we don't know the state of the
	// FP/MMX registers (it could have been changed by the previous filter that
	// used avstp) so we explicitely switch to the FP registers as we're going
	// to use them.
	_mm_empty ();
#endif

	const TaskRsz &	tr = tr_cell._val;
	const TaskRszGlobal &	trg = *(tr._glob_data_ptr);
	assert (trg._this_ptr == this);

	ResizeData *		rd_ptr = 0;
	if (_buffer_flag)
	{
		assert (_factory_aptr.get () != 0);

		rd_ptr = _pool.take_obj ();
		if (rd_ptr == 0)
		{
			throw std::runtime_error (
				"Dither_resize16: Cannot allocate buffer memory."
			);
		}
	}

	int				cur_buf        = 0;
	int				stride_buf [2] = { 0, 0 };
	Dir				cur_dir        = Dir_V;

	// Size of the current tile in the tile's coordinates,
	// not the original picture coordinates.
	int				cur_size [Dir_NBR_ELT] =
	{
		tr._src_end [Dir_H] - tr._src_beg [Dir_H],
		tr._src_end [Dir_V] - tr._src_beg [Dir_V]
	};

	for (int pass = 0; pass < _nbr_passes; ++pass)
	{
		switch (_roadmap [pass])
		{
		case	PassType_RESIZE:

			if (pass == 0)
			{
				assert (cur_dir == Dir_V);

				const int		offset_src = trg._offset_crop + tr._src_beg [Dir_H];
				Scaler::PtrStack16Const::Type	src_ptr =
				{
						trg._src_msb_ptr + offset_src,
						trg._src_lsb_ptr + offset_src
				};

				// Src = MSB/LSB, dst = float
				if (has_buf_dst (pass))
				{
					assert (_buffer_flag);

					stride_buf [cur_buf] = (cur_size [Dir_H] + Scaler::SRC_ALIGN - 1) & -Scaler::SRC_ALIGN;
					assert (tr._work_dst [cur_dir] * stride_buf [cur_buf] <= MAX_BUF_SIZE);
					_scaler_aptr [cur_dir]->process_plane (
						rd_ptr->use_buf (cur_buf),
						src_ptr,
						stride_buf [cur_buf],
						trg._stride_src,
						cur_size [Dir_H],
						tr._dst_beg [cur_dir], tr._dst_beg [cur_dir] + tr._work_dst [cur_dir]
					);
				}

				// Direct vertical scaling (no buffer)
				else
				{
					assert (! _buffer_flag);

					const int		offset_dst = tr._dst_beg [Dir_V] * trg._stride_dst + tr._dst_beg [Dir_H];
					Scaler::PtrStack16::Type	dst_ptr =
					{
						trg._dst_msb_ptr + offset_dst,
						trg._dst_lsb_ptr + offset_dst
					};

					_scaler_aptr [cur_dir]->process_plane (
						dst_ptr,
						src_ptr,
						trg._stride_dst,
						trg._stride_src,
						cur_size [Dir_H],
						tr._dst_beg [cur_dir], tr._dst_beg [cur_dir] + tr._work_dst [cur_dir]
					);
				}
			}

			else
			{
				assert (_buffer_flag);

				const int		offset_src = -tr._src_beg [cur_dir] * stride_buf [cur_buf];

				// Src = float, dst = float
				if (has_buf_dst (pass))
				{
					stride_buf [1 - cur_buf] = stride_buf [cur_buf];
					assert (tr._work_dst [cur_dir] * stride_buf [1 - cur_buf] <= MAX_BUF_SIZE);
					_scaler_aptr [cur_dir]->process_plane (
						rd_ptr->use_buf (1 - cur_buf),
						rd_ptr->use_buf (    cur_buf) + offset_src,
						stride_buf [1 - cur_buf],
						stride_buf [    cur_buf],
						cur_size [Dir_H],
						tr._dst_beg [cur_dir], tr._dst_beg [cur_dir] + tr._work_dst [cur_dir]
					);
					cur_buf = 1 - cur_buf;
				}

				// Src = float, dst = MSB/LSB
				else
				{
					assert (cur_dir == Dir_V);

					const int		offset_dst = tr._dst_beg [Dir_V] * trg._stride_dst + tr._dst_beg [Dir_H];
					Scaler::PtrStack16::Type	dst_ptr =
					{
						trg._dst_msb_ptr + offset_dst,
						trg._dst_lsb_ptr + offset_dst
					};

					_scaler_aptr [cur_dir]->process_plane (
						dst_ptr,
						rd_ptr->use_buf (cur_buf) + offset_src,
						trg._stride_dst,
						stride_buf [cur_buf],
						cur_size [Dir_H],
						tr._dst_beg [cur_dir], tr._dst_beg [cur_dir] + tr._work_dst [cur_dir]
					);
				}
			}

			cur_size [Dir_V] = tr._work_dst [cur_dir];
			break;

		case	PassType_TRANSPOSE:
			// 1st pass: converts input in floating point
			if (! has_buf_src (pass))
			{
				assert (cur_dir == Dir_V);

				stride_buf [cur_buf] =
					(cur_size [Dir_H] + Scaler::SRC_ALIGN - 1) & -Scaler::SRC_ALIGN;
				const int		offset_src =
					  trg._offset_crop
					+ tr._src_beg [Dir_V] * trg._stride_src + tr._src_beg [Dir_H];

				assert (cur_size [Dir_V] * stride_buf [cur_buf] <= MAX_BUF_SIZE);
				(this->*_conv_lsb_msb_to_float_ptr) (
					rd_ptr->use_buf (cur_buf),
					trg._src_msb_ptr + offset_src,
					trg._src_lsb_ptr + offset_src,
					cur_size [Dir_H], cur_size [Dir_V],
					stride_buf [cur_buf],
					trg._stride_src
				);
			}

			// Transposition
			stride_buf [1 - cur_buf] = (cur_size [Dir_V] + Scaler::SRC_ALIGN - 1) & -Scaler::SRC_ALIGN;
			assert (cur_size [Dir_H] * stride_buf [1 - cur_buf] <= MAX_BUF_SIZE);
			transpose (
				rd_ptr->use_buf (1 - cur_buf),
				rd_ptr->use_buf (    cur_buf),
				cur_size [Dir_H], cur_size [Dir_V],
				stride_buf [1 - cur_buf],
				stride_buf [    cur_buf]
			);

			cur_dir = (cur_dir == Dir_V) ? Dir_H : Dir_V;
			std::swap (cur_size [Dir_H], cur_size [Dir_V]);

			cur_buf = 1 - cur_buf;

			// Last pass: converts output to MSB/LSB
			if (! has_buf_dst (pass))
			{
				assert (cur_dir == Dir_V);

				const int		offset_dst = tr._dst_beg [Dir_V] * trg._stride_dst + tr._dst_beg [Dir_H];
				(this->*_conv_float_to_lsb_msb_ptr) (
					trg._dst_msb_ptr + offset_dst,
					trg._dst_lsb_ptr + offset_dst,
					rd_ptr->use_buf (cur_buf),
					tr._work_dst [Dir_H], tr._work_dst [Dir_V],
					trg._stride_dst,
					stride_buf [cur_buf]
				);
			}
			break;

		case	PassType_NONE:
			// Nothing
			break;

		default:
			assert (false);
		}	// switch _roadmap [pass]
	}	// for pass

	if (rd_ptr != 0)
	{
		_pool.return_obj (*rd_ptr);
		rd_ptr = 0;
	}

	_task_rsz_pool.return_cell (tr_cell);
}



void	FilterResize::conv_lsb_msb_to_float_cpp (float *dst_ptr, const uint8_t *src_msb_ptr, const uint8_t *src_lsb_ptr, int w, int h, int stride_dst, int stride_src)
{
	assert (src_msb_ptr != 0);
	assert (src_lsb_ptr != 0);
	assert (w > 0);
	assert (h > 0);
	assert (stride_src > 0);
	assert (dst_ptr != 0);
	assert (stride_dst > 0);

	for (int y = 0; y < h; ++y)
	{
		for (int x = 0; x < w; ++x)
		{
			dst_ptr [x] = float ((src_msb_ptr [x] << 8) + src_lsb_ptr [x]);
		}

		src_msb_ptr += stride_src;
		src_lsb_ptr += stride_src;
		dst_ptr     += stride_dst;
	}
}



void	FilterResize::conv_lsb_msb_to_float_sse (float *dst_ptr, const uint8_t *src_msb_ptr, const uint8_t *src_lsb_ptr, int w, int h, int stride_dst, int stride_src)
{
	assert (src_msb_ptr != 0);
	assert (src_lsb_ptr != 0);
	assert (w > 0);
	assert (h > 0);
	assert (stride_src > 0);
	assert (dst_ptr != 0);
	assert (stride_dst > 0);

	const __m128i	zero    = _mm_setzero_si128 ();

	for (int y = 0; y < h; ++y)
	{
		for (int x = 0; x < w; x += 8)
		{
			const __m128i	src = ToolsSse2::loada_8_16ml (
				src_msb_ptr + x,
				src_lsb_ptr + x
			);
			const __m128i	src_03   = _mm_unpacklo_epi16 (src, zero);
			const __m128i	src_47   = _mm_unpackhi_epi16 (src, zero);
			const __m128	src_f_03 = _mm_cvtepi32_ps (src_03);
			const __m128	src_f_47 = _mm_cvtepi32_ps (src_47);
			_mm_store_ps (dst_ptr + x    , src_f_03);
			_mm_store_ps (dst_ptr + x + 4, src_f_47);
		}

		src_msb_ptr += stride_src;
		src_lsb_ptr += stride_src;
		dst_ptr     += stride_dst;
	}
}



void	FilterResize::conv_float_to_lsb_msb_cpp (uint8_t *dst_msb_ptr, uint8_t *dst_lsb_ptr, const float *src_ptr, int w, int h, int stride_dst, int stride_src)
{
	assert (src_ptr != 0);
	assert (w > 0);
	assert (h > 0);
	assert (stride_src > 0);
	assert (dst_msb_ptr != 0);
	assert (dst_lsb_ptr != 0);
	assert (stride_dst > 0);

	for (int y = 0; y < h; ++y)
	{
		for (int x = 0; x < w; ++x)
		{
			const int		val = limit (conv_int_fast (src_ptr [x]), 0, 65535);
			dst_msb_ptr [x] = uint8_t (val >> 8);
			dst_lsb_ptr [x] = uint8_t (val     );
		}

		src_ptr     += stride_src;
		dst_msb_ptr += stride_dst;
		dst_lsb_ptr += stride_dst;
	}
}



void	FilterResize::conv_float_to_lsb_msb_sse (uint8_t *dst_msb_ptr, uint8_t *dst_lsb_ptr, const float *src_ptr, int w, int h, int stride_dst, int stride_src)
{
	assert (src_ptr != 0);
	assert (w > 0);
	assert (h > 0);
	assert (stride_src > 0);
	assert (dst_msb_ptr != 0);
	assert (dst_lsb_ptr != 0);
	assert (stride_dst > 0);

	const __m128i	mask_lsb = _mm_set1_epi16 (0x00FF);
	const __m128i	sign_bit = _mm_set1_epi16 (-0x8000);
	const __m128	offset   = _mm_set1_ps (32768);

	const int		w8 = w & -8;
	const int		w7 = w - w8;
	const __m128i	mask_store = ToolsSse2::create_store_mask (w7);

	for (int y = 0; y < h; ++y)
	{
		for (int x = 0; x < w8; x += 8)
		{
			__m128			val_03_f = _mm_load_ps (src_ptr + x    );
			__m128			val_47_f = _mm_load_ps (src_ptr + x + 4);
			val_03_f = _mm_sub_ps (val_03_f, offset);
			val_47_f = _mm_sub_ps (val_47_f, offset);

			const __m128i	val_03 = _mm_cvtps_epi32 (val_03_f);
			const __m128i	val_47 = _mm_cvtps_epi32 (val_47_f);

			__m128i			val = _mm_packs_epi32 (val_03, val_47);
			val = _mm_xor_si128 (val, sign_bit);

			ToolsSse2::storea_8_16ml (
				dst_msb_ptr + x,
				dst_lsb_ptr + x,
				val,
				mask_lsb
			);
		}

		if (w7 > 0)
		{
			__m128			val_03_f = _mm_load_ps (src_ptr + w8    );
			__m128			val_47_f = _mm_load_ps (src_ptr + w8 + 4);
			val_03_f = _mm_sub_ps (val_03_f, offset);
			val_47_f = _mm_sub_ps (val_47_f, offset);

			const __m128i	val_03 = _mm_cvtps_epi32 (val_03_f);
			const __m128i	val_47 = _mm_cvtps_epi32 (val_47_f);

			__m128i			val = _mm_packs_epi32 (val_03, val_47);
			val = _mm_xor_si128 (val, sign_bit);

			ToolsSse2::storea_8_16ml_partial (
				dst_msb_ptr + w8,
				dst_lsb_ptr + w8,
				val,
				mask_lsb,
				mask_store
			);
		}

		src_ptr     += stride_src;
		dst_msb_ptr += stride_dst;
		dst_lsb_ptr += stride_dst;
	}
}



// w and h are related to the source.
void	FilterResize::transpose (float *dst_ptr, const float *src_ptr, int w, int h, int stride_dst, int stride_src)
{
	assert (src_ptr != 0);
	assert (w > 0);
	assert (h > 0);
	assert (stride_src > 0);
	assert (dst_ptr != 0);
	assert (stride_dst > 0);

	for (int y = 0; y < h; ++y)
	{
		float *			dst_2_ptr = dst_ptr + y;

		for (int x = 0; x < w; ++x)
		{
			*dst_2_ptr = src_ptr [x];
			dst_2_ptr += stride_dst;
		}

		src_ptr += stride_src;
	}
}



// This function looks not used anymore...
bool	FilterResize::is_kernel_neutral (Dir dir) const
{
	const ContFirInterface &	kernel = *(_kernel_ptr_arr [dir]);

	const double	center = kernel.get_val (0);
	bool				neutral_flag = fabs (center) > 1e-3;
	if (neutral_flag)
	{
		const int		m = ceil_int (kernel.get_support ());
		double			sum = 0;
		for (int k = 1; k <= m && neutral_flag; ++k)
		{
			const double	vn = kernel.get_val (-k);
			const double	vp = kernel.get_val (+k);
			sum += fabs (vn) + fabs (vp);
		}
		neutral_flag = (sum / center < 1e-3);
	}

	return (neutral_flag);
}



bool	FilterResize::has_buf_src (int pass) const
{
	assert (pass >= 0);
	assert (pass < _nbr_passes);

	return (pass > 0);
}



bool	FilterResize::has_buf_dst (int pass) const
{
	assert (pass >= 0);
	assert (pass < _nbr_passes);

	return (pass < _nbr_passes - 1);
}



void	FilterResize::compute_req_src_tile_size (int &tw, int &th, int dw, int dh) const
{
	assert (_buffer_flag);
	assert (_nbr_passes > 0);
	assert (&tw != 0);
	assert (&th != 0);
	assert (dw > 0);
	assert (dh > 0);

	Dir				cur_dir = Dir_V;
	tw = dw;
	th = dh;

	for (int pass = MAX_NBR_PASSES - 1; pass >= 0; --pass)
	{
		switch (_roadmap [pass])
		{
		case	PassType_RESIZE:
			th = Scaler::eval_lower_bound_of_src_tile_height (
				th,
				_dst_size [cur_dir],
				_win_size [cur_dir],
				*(_kernel_ptr_arr [cur_dir]),
				_kernel_scale [cur_dir],
				_src_size [cur_dir]
			);
			break;

		case	PassType_TRANSPOSE:
			std::swap (tw, th);
			cur_dir = (cur_dir == Dir_V) ? Dir_H : Dir_V;
			break;

		case	PassType_NONE:
			// Nothing
			break;

		default:
			assert (false);
			break;
		}
	}

	assert (cur_dir == Dir_V);
}



void	FilterResize::redirect_task_resize (avstp_TaskDispatcher *dispatcher_ptr, void *data_ptr)
{
	TaskRszCell *	trc_ptr = reinterpret_cast <TaskRszCell *> (data_ptr);
	FilterResize *	this_ptr = trc_ptr->_val._glob_data_ptr->_this_ptr;

	this_ptr->process_tile (*trc_ptr);
}



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
