/*****************************************************************************

        Scaler.cpp
        Author: Laurent de Soras, 2011

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if defined (_MSC_VER)
	#pragma warning (1 : 4130 4223 4705 4706)
	#pragma warning (4 : 4355 4786 4800)
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"ContFirInterface.h"
#include	"fnc.h"
#include	"Scaler.h"
#include	"ToolsSse2.h"

#include	<emmintrin.h>

#include	<cassert>



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



Scaler::Scaler (int src_height, int dst_height, double win_top, double win_height, ContFirInterface &kernel_fnc, double kernel_scale, bool norm_flag, double norm_val, double center_pos, bool sse_flag)
:	_src_height (src_height)
,	_dst_height (dst_height)
,	_win_top (win_top)
,	_win_height (win_height)
,	_kernel_scale (kernel_scale)
,	_kernel_fnc (kernel_fnc)
,	_kernel_info_arr (dst_height)
,	_coef_arr ()
,	_process_plane_f32_f32_ptr (&ThisType::process_plane_cpp <ProxyRwCpp <float>, ProxyRwCpp <float> >)
,	_process_plane_f32_s16_ptr (&ThisType::process_plane_cpp <ProxyRwCpp <float>, ProxyRwCpp <uint8_t> >)
,	_process_plane_s16_f32_ptr (&ThisType::process_plane_cpp <ProxyRwCpp <uint8_t>, ProxyRwCpp <float> >)
,	_process_plane_s16_s16_ptr (&ThisType::process_plane_cpp <ProxyRwCpp <uint8_t>, ProxyRwCpp <uint8_t> >)
,	_norm_flag (norm_flag)
,	_norm_val (norm_val)
,	_center_pos (center_pos)
,	_fir_len (0)
{
	assert (src_height > 0);
	assert (dst_height > 0);
	assert (win_height > 0);
	assert (kernel_scale > 0);
	assert (&kernel_fnc != 0);

	if (sse_flag)
	{
		_process_plane_f32_f32_ptr = &ThisType::process_plane_sse <ProxyRwSse <float>, ProxyRwSse <float> >;
		_process_plane_f32_s16_ptr = &ThisType::process_plane_sse <ProxyRwSse <float>, ProxyRwSse <uint8_t> >;
		_process_plane_s16_f32_ptr = &ThisType::process_plane_sse <ProxyRwSse <uint8_t>, ProxyRwSse <float> >;
		_process_plane_s16_s16_ptr = &ThisType::process_plane_sse <ProxyRwSse <uint8_t>, ProxyRwSse <uint8_t> >;
	}

	build_scale_data ();
}



void	Scaler::get_src_boundaries (int &y_src_beg, int &y_src_end, int y_dst_beg, int y_dst_end) const
{
	assert (&y_src_beg != 0);
	assert (&y_src_end != 0);
	assert (y_dst_beg >= 0);
	assert (y_dst_beg < y_dst_end);
	assert (y_dst_end <= _dst_height);

	const KernelInfo &	ki_beg = _kernel_info_arr [y_dst_beg    ];
	const KernelInfo &	ki_end = _kernel_info_arr [y_dst_end - 1];

	y_src_beg = ki_beg._start_line;
	y_src_end = ki_end._start_line + ki_end._kernel_size;

	assert (y_src_beg >= 0);
	assert (y_src_beg < y_src_end);
	assert (y_src_end <= _src_height);
}



int	Scaler::get_fir_len () const
{
	return (_fir_len);
}



// src_ptr is the top-left corner of the full source frame
// dst_ptr is the top-left corner of the destination tile
void	Scaler::process_plane (PtrFloat::Type dst_ptr, PtrFloatConst::Type src_ptr, int dst_stride, int src_stride, int width, int y_dst_beg, int y_dst_end) const
{
	(this->*_process_plane_f32_f32_ptr) (
		dst_ptr, src_ptr,
		dst_stride, src_stride,
		width, y_dst_beg, y_dst_end
	);
}

void	Scaler::process_plane (PtrFloat::Type dst_ptr, PtrStack16Const::Type src_ptr, int dst_stride, int src_stride, int width, int y_dst_beg, int y_dst_end) const
{
	(this->*_process_plane_f32_s16_ptr) (
		dst_ptr, src_ptr,
		dst_stride, src_stride,
		width, y_dst_beg, y_dst_end
	);
}

void	Scaler::process_plane (PtrStack16::Type dst_ptr, PtrFloatConst::Type src_ptr, int dst_stride, int src_stride, int width, int y_dst_beg, int y_dst_end) const
{
	(this->*_process_plane_s16_f32_ptr) (
		dst_ptr, src_ptr,
		dst_stride, src_stride,
		width, y_dst_beg, y_dst_end
	);
}

void	Scaler::process_plane (PtrStack16::Type dst_ptr, PtrStack16Const::Type src_ptr, int dst_stride, int src_stride, int width, int y_dst_beg, int y_dst_end) const
{
	(this->*_process_plane_s16_s16_ptr) (
		dst_ptr, src_ptr,
		dst_stride, src_stride,
		width, y_dst_beg, y_dst_end
	);
}



void	Scaler::eval_req_src_area (int &work_top, int &work_height, int src_height, int dst_height, double win_top, double win_height, ContFirInterface &kernel_fnc, double kernel_scale, double center_pos)
{
	assert (&work_top != 0);
	assert (&work_height != 0);
	assert (src_height > 0);
	assert (dst_height > 0);
	assert (win_height > 0);
	assert (kernel_scale > 0);
	assert (&kernel_fnc != 0);

	const BasicInfo	bi (
		src_height, dst_height, win_top, win_height,
		kernel_fnc, kernel_scale, center_pos
	);

	double			pos_flt = bi._src_pos + bi._support - bi._fir_len + 1;

	int				src_pos_beg = floor_int (pos_flt);
	src_pos_beg = limit (src_pos_beg, 0, src_height - 1);

	pos_flt    += (dst_height - 1) * bi._src_step;

	int				src_pos_end = floor_int (pos_flt + bi._fir_len);
	src_pos_end = limit (src_pos_end, 1, src_height);

	work_top    = src_pos_beg;
	work_height = src_pos_end - src_pos_beg;
}



int	Scaler::eval_lower_bound_of_dst_tile_height (int tile_height_src, int dst_height, double win_height, ContFirInterface &kernel_fnc, double kernel_scale, int src_height)
{
	assert (tile_height_src > 0);
	assert (dst_height > 0);
	assert (win_height > 0);
//	assert (win_height >= tile_height_src);
	assert (kernel_scale > 0);
	assert (&kernel_fnc != 0);
	assert (src_height > 0);

	int				tile_height_dst = 0;
	if (tile_height_src >= src_height)
	{
		tile_height_dst = dst_height;
	}

	else
	{
		const BasicInfo	bi (
			ceil_int (win_height), dst_height, 0, win_height,
			kernel_fnc, kernel_scale, 0
		);

		// Technically this should be "- bi._fir_len + 1" but we have use
		// the minimum possible value.
		const int		real_height_src = tile_height_src - bi._fir_len;
		const double	tile_height_dst_flt = real_height_src / bi._src_step;
		tile_height_dst = floor_int (tile_height_dst_flt);
	}

	return (tile_height_dst);
}



int	Scaler::eval_lower_bound_of_src_tile_height (int tile_height_dst, int dst_height, double win_height, ContFirInterface &kernel_fnc, double kernel_scale, int src_height)
{
	assert (tile_height_dst > 0);
	assert (dst_height > 0);
//	assert (dst_height >= tile_height_dst);
	assert (win_height > 0);
	assert (kernel_scale > 0);
	assert (&kernel_fnc != 0);

	const BasicInfo	bi (
		ceil_int (win_height), dst_height, 0, win_height,
		kernel_fnc, kernel_scale, 0
	);

	const int		h = ceil_int (tile_height_dst * bi._src_step);
	int				tile_height_src = h + bi._fir_len - 1;

	const int		y_src_beg = limit (
		floor_int (bi._src_pos + bi._support) - bi._fir_len + 1,
		0,
		src_height - 1
	);
	const int		y_src_end = limit (
		floor_int (bi._src_pos + bi._src_step * dst_height + bi._support) + 1,
		1,
		src_height
	);
	const int		max_src_height = y_src_end - y_src_beg;

	tile_height_src = std::min (tile_height_src, max_src_height);

	return (tile_height_src);
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



template <class DST, class SRC>
void	Scaler::process_plane_cpp (typename DST::Ptr::Type dst_ptr, typename SRC::PtrConst::Type src_ptr, int dst_stride, int src_stride, int width, int y_dst_beg, int y_dst_end) const
{
	assert (DST::Ptr::check_ptr (dst_ptr));
	assert (SRC::PtrConst::check_ptr (src_ptr));
	assert (dst_stride != 0);
	assert (width > 0);
	assert (y_dst_beg >= 0);
	assert (y_dst_beg < y_dst_end);
	assert (y_dst_end <= _dst_height);

	for (int y = y_dst_beg; y < y_dst_end; ++y)
	{
		const KernelInfo &	kernel_info = _kernel_info_arr [y];
		const int		kernel_size   = kernel_info._kernel_size;
		const float *	coef_base_ptr = &_coef_arr [kernel_info._coef_index];
		const int		ofs_y         = kernel_info._start_line;

		SRC::PtrConst::Type	col_src_ptr = src_ptr;
		SRC::PtrConst::jump (col_src_ptr, src_stride * ofs_y);
		DST::Ptr::Type			col_dst_ptr = dst_ptr;

		for (int x = 0; x < width; x += 2)
		{
			float			sum0 = 0;
			float			sum1 = 0;

			SRC::PtrConst::Type	pix_ptr = col_src_ptr;
			for (int k = 0; k < kernel_size; ++k)
			{
				const float		coef = coef_base_ptr [k];
				float		src0;
				float		src1;
				SRC::read (pix_ptr, src0, src1);
				sum0 += src0 * coef;
				sum1 += src1 * coef;

				SRC::PtrConst::jump (pix_ptr, src_stride);
			}

			DST::write (col_dst_ptr, sum0, sum1);

			DST::Ptr::jump (col_dst_ptr, 2);
			SRC::PtrConst::jump (col_src_ptr, 2);
		}

		DST::Ptr::jump (dst_ptr, dst_stride);
	}
}



template <class DST, class SRC>
void	Scaler::process_plane_sse (typename DST::Ptr::Type dst_ptr, typename SRC::PtrConst::Type src_ptr, int dst_stride, int src_stride, int width, int y_dst_beg, int y_dst_end) const
{
	assert (DST::Ptr::check_ptr (dst_ptr, DST::ALIGN));
	assert (SRC::PtrConst::check_ptr (src_ptr, SRC::ALIGN));
	// mod4 is enough to guarantee alignment, but since we process pairs of
	// vectors and write_partial() is not different from write() with float
	// data (overwriting all the 32 bytes), we must take extra-care not to
	// overflow from the output buffer.
	assert ((dst_stride & 7) == 0);	
	assert ((src_stride & 3) == 0);
	assert (width > 0);
	assert (y_dst_beg >= 0);
	assert (y_dst_beg < y_dst_end);
	assert (y_dst_end <= _dst_height);

	const __m128i	zero = _mm_setzero_si128 ();
	const __m128i	mask_lsb = _mm_set1_epi16 (0x00FF);
	const __m128i	sign_bit = _mm_set1_epi16 (-0x8000);
	const __m128	offset   = _mm_set1_ps (32768);

	const int		w8 = width & -8;
	const int		w7 = width - w8;
	const __m128i	mask_store = ToolsSse2::create_store_mask (w7);

	for (int y = y_dst_beg; y < y_dst_end; ++y)
	{
		const KernelInfo &	kernel_info = _kernel_info_arr [y];
		const int		kernel_size = kernel_info._kernel_size;
		const float *	coef_base_ptr = &_coef_arr [kernel_info._coef_index];
		const int		ofs_y       = kernel_info._start_line;

		SRC::PtrConst::Type	col_src_ptr = src_ptr;
		SRC::PtrConst::jump (col_src_ptr, src_stride * ofs_y);
		DST::Ptr::Type			col_dst_ptr = dst_ptr;

		__m128			sum0;
		__m128			sum1;

		for (int x = 0; x < w8; x += 8)
		{
			SRC::PtrConst::Type	pix_ptr = col_src_ptr;

			process_vect_sse <SRC> (
				sum0, sum1, kernel_size, coef_base_ptr, pix_ptr, zero, src_stride
			);
			DST::write (col_dst_ptr, sum0, sum1, mask_lsb, sign_bit, offset);

			DST::Ptr::jump (col_dst_ptr, 8);
			SRC::PtrConst::jump (col_src_ptr, 8);
		}

		if (w7 > 0)
		{
			SRC::PtrConst::Type	pix_ptr = col_src_ptr;

			process_vect_sse <SRC> (
				sum0, sum1, kernel_size, coef_base_ptr, pix_ptr, zero, src_stride
			);
			DST::write_partial (
				col_dst_ptr, sum0, sum1, mask_lsb, sign_bit, offset, mask_store
			);
		}

		DST::Ptr::jump (dst_ptr, dst_stride);
	}
}



template <class SRC>
void	Scaler::process_vect_sse (__m128 &sum0, __m128 &sum1, int kernel_size, const float *coef_base_ptr, typename SRC::PtrConst::Type pix_ptr, const __m128i &zero, int src_stride)
{
	sum0 = _mm_setzero_ps ();
	sum1 = _mm_setzero_ps ();

	for (int k = 0; k < kernel_size; ++k)
	{
		__m128			coef = _mm_load_ss (coef_base_ptr + k);
		coef = _mm_shuffle_ps (coef, coef, 0);
		__m128			src0;
		__m128			src1;
		SRC::read (pix_ptr, src0, src1, zero);
		const __m128	val0 = _mm_mul_ps (src0, coef);
		const __m128	val1 = _mm_mul_ps (src1, coef);
		sum0 = _mm_add_ps (sum0, val0);
		sum1 = _mm_add_ps (sum1, val1);

		SRC::PtrConst::jump (pix_ptr, src_stride);
	}
}



void	Scaler::build_scale_data ()
{
	_coef_arr.clear ();

	BasicInfo		bi (
		_src_height, _dst_height, _win_top, _win_height,
		_kernel_fnc, _kernel_scale, _center_pos
	);

	_fir_len = bi._fir_len;

	std::vector <double>	coef_tmp;
	const int		last_line = _src_height - 1;

	for (int y = 0; y < _dst_height; ++y)
	{
		// First pass: collects the coefficients and compute their sum to
		// normalize the impulse energy.
		coef_tmp.clear ();

		int				src_pos_beg =
			floor_int (bi._src_pos + bi._support) - bi._fir_len + 1;
		double			sum = 0;
		for (int k = 0; k < bi._fir_len; ++k)
		{
			const int		p          = src_pos_beg + k;
			const double	pos_in_fir = (bi._src_pos - p) * bi._imp_step;
			const double	val        = _kernel_fnc.get_val (pos_in_fir);

			coef_tmp.push_back (val);
			sum += val;
		}

		double			amp = 1;
		if (_norm_flag)
		{
			double			n = sum;
			if (_norm_val > 0)
			{
				n = _norm_val;
			}
			assert (n != 0);
			amp = 1.0 / fabs (n);
		}

		// Second pass: builds the actual FIR, handling picture edge conditions.
		KernelInfo &	info = _kernel_info_arr [y];
		double			accu = 0;
		info._kernel_size   = 0;
		info._coef_index    = int (_coef_arr.size ());
		info._start_line    = limit (src_pos_beg, 0, last_line);
		for (int k = 0; k < bi._fir_len; ++k)
		{
			const int		p   = src_pos_beg + k;
			const double	val = coef_tmp [k] * amp;
			accu += val;
			if (p >= 0 && p < last_line)
			{
				++ info._kernel_size;
				_coef_arr.push_back (float (accu));
				accu = 0;
			}
		}

		// Case 1: We've gone past the bottom of the source and haven't stored
		//         the last line data yet
		// Case 2: We even haven't reached any valid line.
		if (accu != 0 || info._kernel_size == 0)
		{
			++ info._kernel_size;
			_coef_arr.push_back (float (accu));
		}

		// Next line
		bi._src_pos += bi._src_step;
	}
}



Scaler::BasicInfo::BasicInfo (int src_height, int dst_height, double win_top, double win_height, ContFirInterface &kernel_fnc, double kernel_scale, double center_pos)
{
	assert (src_height > 0);
	assert (dst_height > 0);
	assert (win_height > 0);
	assert (kernel_scale > 0);
	assert (&kernel_fnc != 0);

	// step is the distance between the two source pixels corresponding
	// to two adjacent destination pixels
	_src_step = win_height / double (dst_height);

	// Size of a filter unit step (distance between 2 zero-crossing for a sinc)
	_zc_size  = std::max (_src_step, 1.0) / kernel_scale;

	// This is the correspounding distance in the impulse.
	_imp_step = 1.0 / _zc_size;

	// Number of source pixels covered by the FIR
	_support  = kernel_fnc.get_support () * _zc_size;
	_fir_len  = std::max (ceil_int (_support * 2), 1);

	_src_pos  = win_top;

	// Introduces an offset because samples are located at the center of the
	// pixels, not on their boundaries. Excepted for pointresize.
	if (_support > 0)
	{
		_src_pos += (_src_step - 1) * center_pos;
	}
	else
	{
		// Not really required, but gives the same behaviour as
		// AviSynth's PointResize().
		_support = 0.0001;
	}
}



void	Scaler::ProxyRwCpp <float>::read (const PtrConst::Type &ptr, float &src0, float &src1)
{
	src0 = ptr [0];
	src1 = ptr [1];
}

void	Scaler::ProxyRwCpp <float>::write (const Ptr::Type &ptr, const float &src0, const float &src1)
{
	ptr [0] = src0;
	ptr [1] = src1;
}

void	Scaler::ProxyRwCpp <uint8_t>::read (const PtrConst::Type &ptr, float &src0, float &src1)
{
	src0 = float ((ptr._msb_ptr [0] << 8) | ptr._lsb_ptr [0]);
	src1 = float ((ptr._msb_ptr [1] << 8) | ptr._lsb_ptr [1]);
}

void	Scaler::ProxyRwCpp <uint8_t>::write (const Ptr::Type &ptr, const float &src0, const float &src1)
{
	const int		v0 = limit (conv_int_fast (src0), 0, 65535);
	const int		v1 = limit (conv_int_fast (src1), 0, 65535);

	ptr._msb_ptr [0] = uint8_t (v0 >> 8);
	ptr._lsb_ptr [0] = uint8_t (v0     );
	ptr._msb_ptr [1] = uint8_t (v1 >> 8);
	ptr._lsb_ptr [1] = uint8_t (v1     );
}



void	Scaler::ProxyRwSse <float>::read (const PtrConst::Type &ptr, __m128 &src0, __m128 &src1, const __m128i &/*zero*/)
{
	src0 = _mm_load_ps (ptr    );
	src1 = _mm_load_ps (ptr + 4);
}

void	Scaler::ProxyRwSse <float>::write (const Ptr::Type &ptr, const __m128 &src0, const __m128 &src1, const __m128i &/*mask_lsb*/, const __m128i &/*sign_bit*/, const __m128 &/*offset*/)
{
	_mm_store_ps (ptr,     src0);
	_mm_store_ps (ptr + 4, src1);
}

void	Scaler::ProxyRwSse <float>::write_partial (const Ptr::Type &ptr, const __m128 &src0, const __m128 &src1, const __m128i &/*mask_lsb*/, const __m128i &/*sign_bit*/, const __m128 &/*offset*/, const __m128i &/*mask_store*/)
{
	_mm_store_ps (ptr,     src0);
	_mm_store_ps (ptr + 4, src1);
}

// const __m128i	zero = _mm_setzero_si128 ();
void	Scaler::ProxyRwSse <uint8_t>::read (const PtrConst::Type &ptr, __m128 &src0, __m128 &src1, const __m128i &zero)
{
	const __m128i	src = ToolsSse2::loada_8_16ml (ptr._msb_ptr, ptr._lsb_ptr);
	const __m128i	src_03   = _mm_unpacklo_epi16 (src, zero);
	const __m128i	src_47   = _mm_unpackhi_epi16 (src, zero);
	src0 = _mm_cvtepi32_ps (src_03);
	src1 = _mm_cvtepi32_ps (src_47);
}

//	const __m128i	mask_lsb = _mm_set1_epi16 (0x00FF);
//	const __m128i	sign_bit = _mm_set1_epi16 (-0x8000);
//	const __m128	offset   = _mm_set1_ps (32768);
void	Scaler::ProxyRwSse <uint8_t>::write (const Ptr::Type &ptr, const __m128 &src0, const __m128 &src1, const __m128i &mask_lsb, const __m128i &sign_bit, const __m128 &offset)
{
	__m128			val_03_f = _mm_sub_ps (src0, offset);
	__m128			val_47_f = _mm_sub_ps (src1, offset);

	const __m128i	val_03 = _mm_cvtps_epi32 (val_03_f);
	const __m128i	val_47 = _mm_cvtps_epi32 (val_47_f);

	__m128i			val = _mm_packs_epi32 (val_03, val_47);
	val = _mm_xor_si128 (val, sign_bit);

	ToolsSse2::storea_8_16ml (
		ptr._msb_ptr,
		ptr._lsb_ptr,
		val,
		mask_lsb
	);
}

void	Scaler::ProxyRwSse <uint8_t>::write_partial (const Ptr::Type &ptr, const __m128 &src0, const __m128 &src1, const __m128i &mask_lsb, const __m128i &sign_bit, const __m128 &offset, const __m128i &mask_store)
{
	__m128			val_03_f = _mm_sub_ps (src0, offset);
	__m128			val_47_f = _mm_sub_ps (src1, offset);

	const __m128i	val_03 = _mm_cvtps_epi32 (val_03_f);
	const __m128i	val_47 = _mm_cvtps_epi32 (val_47_f);

	__m128i			val = _mm_packs_epi32 (val_03, val_47);
	val = _mm_xor_si128 (val, sign_bit);

	ToolsSse2::storea_8_16ml_partial (
		ptr._msb_ptr,
		ptr._lsb_ptr,
		val,
		mask_lsb,
		mask_store
	);
}



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
